import sys
import pygame
import math
import random
import time

class GameObject(object):
    def __init__(self, width, height, screen, level=1):
        self.scr_width_ = width
        self.scr_height_ = height
        self.level_ = level
        self.screen_ = screen
    
    def updateLevel(self, level):
        self.level_ = level

    def inScreen(self):
        return True

    def update(self):
        pass

class ImagedGameObject(GameObject):
    def __init__(self, width, height, screen, img_name,
                 start_x, start_y, level=1):
        self.x_ = start_x
        self.y_ = start_y
        self.img_ = pygame.image.load(img_name)
        self.x_offset_ = self.img_.get_rect().width / 2
        self.y_offset_ = self.img_.get_rect().height / 2
        
        super(ImagedGameObject, self).__init__(width, height, screen, level)


    def checkWindCollide(self, pts):
        ## generate object bounding box
        left = (self.x_ - self.x_offset_) + 1
        right = (self.x_ + self.x_offset_) + 1
        top = self.y_ - self.y_offset_
        bottom = self.y_ + self.y_offset_

        ## pare down points to check
        y_pts = [y for x, y in pts[left:right+1]]
        if len(y_pts) <= 0:
            return False

        ## find the max of the points
        y_max = max(y_pts)

        ## see if the max falls in the bounding obj
        if (y_max >= top and y_max <= self.y_):
            # self.y_ += (self.y_offset_)
            return True
        elif (y_max > self.y_ and y_max <= bottom):
            # self.y_ -= (self.y_offset_)
            return True
            
        return False

    def checkBoundedCollide(self, obj):
        lr_collide = False
        tb_collide = False
        
        self_left = self.x_ - self.x_offset_
        self_right = self.x_ + self.x_offset_
        
        other_right = obj.x_ + obj.x_offset_
        other_left = obj.x_ - obj.x_offset_
        
        if other_right > self_left and other_left < self_right:
            lr_collide = True

        self_top = self.y_ - self.y_offset_
        self_bottom = self.y_ + self.y_offset_

        other_bottom = obj.y_ + obj.y_offset_
        other_top = obj.y_ - obj.y_offset_

        
        if other_top < self_bottom and other_bottom > self_top:
            tb_collide = True


        if lr_collide and tb_collide:
            return True
        else:
            return False
    
                
class Wind(GameObject):
    def __init__(self, width, height, screen, offset, direction, level=1):
        self.wave_points_ = [] ## saved y points
        self.render_points_ = [] ## x-y locs used for the screen position

        ## values for computing the sin wave - ie "wind"
        self.ampl_ = 10
        self.per_ = 2
        self.angle_ = random.randint(0, 360) ## winds start at diff points

        self.offset_ = offset
        self.dir_ = direction  ## negative direction is left
        
        self.gusting_ = None
        self.gust_wait_ = 0
        self.gusting_bool_ = False

        super(Wind, self).__init__(width, height, screen, level)

        
    ## helper function for predictions
    def _next_angle(self, prev_angle):
        return (prev_angle + 1) % 361
    
    def _draw_wave(self):
        ## age off old data
        if (len(self.wave_points_) >= self.scr_width_):
            self.wave_points_ = self.wave_points_[1:]
        
        ## compute yPos
        yPos = self.ampl_ * math.sin(self.per_ * math.radians(self.angle_))
        yPos += self.offset_
            
        ## append new point
        self.wave_points_.append(yPos)

        ## determine the x range to draw the points over; return if
        ## drawlines can't handle the range (ie, range of 1)
        x_render_range = range(0, len(self.wave_points_))

        if self.dir_ < 0:
            x_offset = self.scr_width_ - len(self.wave_points_)
            x_render_range = range(x_offset, self.scr_width_)

        if len(x_render_range) <= 1:
            return

        ## create the x-y coords to render and put them on the screen
        self.render_points_ = zip(x_render_range, self.wave_points_)
        pygame.draw.lines(self.screen_, pygame.Color("red"), False, self.render_points_, 2)
        
    def getPts(self):
        return self.render_points_

    def getDirection(self):
        return self.dir_
    
    def update(self):
        ## we are gusting, see if it's time to stop!
        if self.gusting_:
            if self.gusting_ <= 500:
                self.gusting_ += 1
            elif self.gusting_ > 500:
                self.gusting_ = None
                self.ampl_ = 10
                self.gust_wait_ = 0
                self.gusting_bool_ = False
        ## see if we should start gusting!
        else:
            self.gust_wait_ += 1
            if self.gust_wait_ > 400:
            ## so hacky :(
                chance = random.randint(1, 1000 * self.level_)
                if chance in range(1, (self.level_ + 1)):
                    self.gusting_ = 1
                    self.ampl_ = 30
                    self.gusting_bool_ = True
            
        self._draw_wave()
        self.angle_ = self._next_angle(self.angle_)

## balloons need to follow this algorithm:
## rise with drift
## if you hit wind drift faster (+1) while rising
## if you are told you were hit, explode (maybe)
## extra: handle collissions with other objs
class Balloon(ImagedGameObject):
    ## easy-going balloons have drag in the 4-6 range
    def __init__(self, width, height, screen, level=1):
        x = random.randint(0, 1024) ## TODO: refine
        y = 576 + 5 ## just out of range of the thing
        self.move_ = 0
        self.drag_ = random.randint(4,6)  ## higher numbers are slower
        self.drift_ = random.randint(-1, 1)
        self.extra_drift_ = 0

        super(Balloon, self).__init__(width, height, screen, \
                                      "balloon.png", x, y, level)
        
    def inScreen(self):
        bottom = self.y_ + self.y_offset_
        if bottom <= 0:
            return False
        return True

    def handleCollisions(self, objs):
        self.extra_drift_ = 0
        for obj in objs:
            if obj is self:
                continue
            elif type(obj) is Wind:
                if self.checkWindCollide(obj.getPts()):
                    self.extra_drift_ = 2 * obj.getDirection()
        
    
    def update(self):
        ## movement:
        if self.move_ == 0:
            self.y_ -= 1
            if self.extra_drift_ != 0:
                self.x_ += self.extra_drift_
            else:
                self.x_ -= self.drift_

            
        self.move_ = (self.move_ + 1) % self.drag_

        x = self.x_ - self.x_offset_
        y = self.y_ - self.y_offset_

        self.screen_.blit(self.img_, (x, y))


class Plane(ImagedGameObject):
    def __init__(self, width, height, screen, x, y, level=1):
        self.health_ = 100
        self.orient_ = 0  ## direction

        super(Plane, self).__init__(width, height, screen, \
                                    "plane.png", x, y, level)

    def handleCollisions(self, objs):
        for obj in objs:
            if obj is self:
                continue
            elif type(obj) is Wind:
                if self.checkWindCollide(obj.getPts()):
                    wind_dir = obj.getDirection()                
                    self.x_ += (4 * wind_dir)

                    ## check left bound
                    left_bound = self.x_ - self.x_offset_
                    if left_bound < 0:
                        self.x_ = self.x_offset_

                    ## check right bound
                    right_bound = self.x_ + self.x_offset_
                    if right_bound > self.scr_width_:
                        self.x_ = self.scr_width_ - self.x_offset_
            else:
                if self.checkBoundedCollide(obj):
                    if type(obj) is Balloon:
                        self.health_ -= 1
    

    def _boundsOk(self, delta):
        y = self.y_ + delta

        if delta > 0:
            bottom = y + self.y_offset_
            if bottom >= self.scr_height_:
                return False

        if delta < 0:
            top = y - self.y_offset_
            if top <= 0:
                return False
        return True
    
    def handleKeyPress(self, pressed):
        if pressed[pygame.K_LEFT]:
            pass
        elif pressed[pygame.K_RIGHT]:
            pass
        elif pressed[pygame.K_UP]:
            if self._boundsOk(-1):
                self.y_ -= 1
            ## self.img_ = pygame.transform.rotate(self.img_, 1)
        elif pressed[pygame.K_DOWN]:
            if self._boundsOk(1):
                self.y_ += 1
            ## self.img_ = pygame.transform.rotate(self.img_, -1)


    def update(self):
        ## draw the plane at x,y and rotate it
        x = self.x_ - self.x_offset_
        y = self.y_ - self.y_offset_

        self.screen_.blit(self.img_, (x, y))
    
# set the basics
title = "Love Note"
WIDTH = 1024
HEIGHT = 576
size = [WIDTH, HEIGHT]

# init the screen + engine-y parts
pygame.init()
pygame.display.set_caption(title)
screen = pygame.display.set_mode(size)
font = pygame.font.SysFont("monospace", 20)
in_game = True

bg = pygame.image.load("clouds.png")


wind1 = Wind(WIDTH, HEIGHT, screen, (HEIGHT - (HEIGHT/6)), -1)
wind2 = Wind(WIDTH, HEIGHT, screen, (HEIGHT/6), 1)
plane = Plane(WIDTH, HEIGHT, screen, WIDTH/2, 150)

game_objs = [wind1, wind2, plane]
game_level_info = {1: {'duration': 10,
                       'objs': [(Balloon, 500)]},
                   2: {'duration': 15,
                       'objs': [(Balloon, 1000)]},
                   3: {'duration': 20,
                       'objs': [(Balloon, 2000)]},
                   4: {'duration': 20,
                       'objs': [(Balloon, 1000)]},
                   5: {'duration': 20,
                       'objs': [(Balloon, 500)]}
                   }
# game event loop
level = 1
tm_prev = time.time()
while True:
    ## check for quit first thing
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.quit()

    if in_game:
        tm_curr = time.time()
        sec_diff = int(tm_curr - tm_prev)
        if sec_diff >= game_level_info[level]['duration']:
            level += 1
            wind1.updateLevel(level)
        
            if (level > len(game_level_info)):
                ## you have beaten the game! show a success screen!
                in_game = False
                
                screen.fill(pygame.Color("black"))
                label = font.render("Success!", 1, (255, 0, 255))
                screen.blit(label, ((1024/2), (576/2)))
            
            ## TODO update the wind!
            tm_prev = tm_curr
            sec_dif = 0

    ## if we are still on a level, run all the update code
    if in_game:
        screen.fill(pygame.Color("blue"))
        screen.blit(bg, (0,0))
        
        ## debug label
        debug_str = "Level " + str(level) + ": " + str(sec_diff) + "s"
        label = font.render(debug_str, 1, (255, 0, 255))
        screen.blit(label, ((1024 - 150), (576-40)))

        ## user update
        pressed = pygame.key.get_pressed()
        plane.handleKeyPress(pressed)

        ## spawn and append list
        objs = game_level_info[level]['objs']
        for ob, prob in objs:
            chance = random.randint(1, prob)
            if chance == 1:
                game_objs.append(ob(WIDTH, HEIGHT, screen))

        to_remove = []
        for idx, obj in enumerate(game_objs):
            if obj.inScreen():
                obj.update()
            else:
                to_remove.append(idx)

        ## clean out expired objects
        to_remove.reverse()
        for idx in to_remove:
            game_objs.pop(idx)
            
        for collidable in game_objs[2:]:
            collidable.handleCollisions(game_objs)

        if plane.health_ <= 0:
            in_game = False
            screen.fill(pygame.Color("black"))
            label = font.render("You died!", 1, (255, 0, 255))
            screen.blit(label, ((1024/2), (576/2)))
            
    pygame.display.update()
